﻿// Copyright (c) Microsoft. All rights reserved.
// Licensed under the MIT license. See LICENSE file in the project root for full license information.

namespace Microsoft.Azure.Amqp.Encoding
{
    using System;
    using System.Collections;
    using System.Collections.Generic;

    /// <summary>
    /// Encodes and decodes AMQP types.
    /// </summary>
    public static class AmqpEncoding
    {
        static Dictionary<Type, EncodingBase> encodingsByType;
        static Dictionary<FormatCode, EncodingBase> encodingsByCode;

        static BooleanEncoding booleanEncoding = new BooleanEncoding();
        static UByteEncoding ubyteEncoding = new UByteEncoding();
        static UShortEncoding ushortEncoding = new UShortEncoding();
        static UIntEncoding uintEncoding = new UIntEncoding();
        static ULongEncoding ulongEncoding = new ULongEncoding();
        static ByteEncoding byteEncoding = new ByteEncoding();
        static ShortEncoding shortEncoding = new ShortEncoding();
        static IntEncoding intEncoding = new IntEncoding();
        static LongEncoding longEncoding = new LongEncoding();
        static FloatEncoding floatEncoding = new FloatEncoding();
        static DoubleEncoding doubleEncoding = new DoubleEncoding();
        static DecimalEncoding decimal128Encoding = new DecimalEncoding();
        static CharEncoding charEncoding = new CharEncoding();
        static TimeStampEncoding timeStampEncoding = new TimeStampEncoding();
        static UuidEncoding uuidEncoding = new UuidEncoding();
        static BinaryEncoding binaryEncoding = new BinaryEncoding();
        static SymbolEncoding symbolEncoding = new SymbolEncoding();
        static StringEncoding stringEncoding = new StringEncoding();
        static ListEncoding listEncoding = new ListEncoding();
        static MapEncoding mapEncoding = new MapEncoding();
        static ArrayEncoding arrayEncoding = new ArrayEncoding();
        static DescribedEncoding describedTypeEncoding = new DescribedEncoding();

        static AmqpEncoding()
        {
            encodingsByType = new Dictionary<Type, EncodingBase>()
            {
                { typeof(bool),             booleanEncoding },
                { typeof(byte),             ubyteEncoding },
                { typeof(ushort),           ushortEncoding },
                { typeof(uint),             uintEncoding },
                { typeof(ulong),            ulongEncoding },
                { typeof(sbyte),            byteEncoding },
                { typeof(short),            shortEncoding },
                { typeof(int),              intEncoding },
                { typeof(long),             longEncoding },
                { typeof(float),            floatEncoding },
                { typeof(double),           doubleEncoding },
                { typeof(decimal),          decimal128Encoding },
                { typeof(char),             charEncoding },
                { typeof(DateTime),         timeStampEncoding },
                { typeof(Guid),             uuidEncoding },
                { typeof(ArraySegment<byte>), binaryEncoding },
                { typeof(AmqpSymbol),       symbolEncoding },
                { typeof(string),           stringEncoding },
                { typeof(AmqpMap),          mapEncoding },
            };

            encodingsByCode = new Dictionary<FormatCode, EncodingBase>()
            {
                { FormatCode.BooleanFalse,  booleanEncoding },
                { FormatCode.BooleanTrue,   booleanEncoding },
                { FormatCode.Boolean,       booleanEncoding },
                { FormatCode.UByte,         ubyteEncoding },
                { FormatCode.UShort,        ushortEncoding },
                { FormatCode.UInt,          uintEncoding },
                { FormatCode.SmallUInt,     uintEncoding },
                { FormatCode.UInt0,         uintEncoding },
                { FormatCode.ULong,         ulongEncoding },
                { FormatCode.SmallULong,    ulongEncoding },
                { FormatCode.ULong0,        ulongEncoding },
                { FormatCode.Byte,          byteEncoding },
                { FormatCode.Short,         shortEncoding },
                { FormatCode.Int,           intEncoding },
                { FormatCode.SmallInt,      intEncoding },
                { FormatCode.Long,          longEncoding },
                { FormatCode.SmallLong,     longEncoding },
                { FormatCode.Float,         floatEncoding },
                { FormatCode.Double,        doubleEncoding },
                { FormatCode.Decimal128,    decimal128Encoding },
                { FormatCode.Char,          charEncoding },
                { FormatCode.TimeStamp,     timeStampEncoding },
                { FormatCode.Uuid,          uuidEncoding },
                { FormatCode.Binary8,       binaryEncoding },
                { FormatCode.Binary32,      binaryEncoding },
                { FormatCode.Symbol8,       symbolEncoding },
                { FormatCode.Symbol32,      symbolEncoding },
                { FormatCode.String8Utf8,   stringEncoding },
                { FormatCode.String32Utf8,  stringEncoding },
                { FormatCode.List0,         listEncoding },
                { FormatCode.List8,         listEncoding },
                { FormatCode.List32,        listEncoding },
                { FormatCode.Map8,          mapEncoding },
                { FormatCode.Map32,         mapEncoding },
                { FormatCode.Array8,        arrayEncoding },
                { FormatCode.Array32,       arrayEncoding },
                { FormatCode.Described,     describedTypeEncoding }
            };
        }

        internal static EncodingBase GetEncoding(object value)
        {
            return GetEncoding(value.GetType());
        }

        /// <summary>
        /// Gets a <see cref="EncodingBase"/> for a given type.
        /// </summary>
        /// <param name="type">The type.</param>
        /// <returns>A <see cref="EncodingBase"/> for the type.</returns>
        internal static EncodingBase GetEncoding(Type type)
        {
            EncodingBase encoding = null;
            if (encodingsByType.TryGetValue(type, out encoding))
            {
                return encoding;
            }
            else if (type.IsArray)
            {
                return arrayEncoding;
            }
            else if (typeof(IList).IsAssignableFrom(type))
            {
                return listEncoding;
            }
            else if (typeof(DescribedType).IsAssignableFrom(type))
            {
                return describedTypeEncoding;
            }

            throw new NotSupportedException(AmqpResources.GetString(AmqpResources.AmqpInvalidType, type.FullName));
        }

        internal static EncodingBase GetEncoding(FormatCode formatCode)
        {
            EncodingBase encoding;
            if (encodingsByCode.TryGetValue(formatCode, out encoding))
            {
                return encoding;
            }

            return null;
        }

        internal static int GetEncodeWidthBySize(int size)
        {
            return size <= byte.MaxValue ? FixedWidth.UByte : FixedWidth.UInt;
        }

        internal static int GetEncodeWidthByCountAndSize(int count, int valueSize)
        {
            return count < byte.MaxValue && valueSize < byte.MaxValue ? FixedWidth.UByte : FixedWidth.UInt;
        }

        internal static FormatCode ReadFormatCode(ByteBuffer buffer)
        {
            byte type = AmqpBitConverter.ReadUByte(buffer);
            byte extType = 0;
            if (FormatCode.HasExtType(type))
            {
                extType = AmqpBitConverter.ReadUByte(buffer);
            }

            return new FormatCode(type, extType);
        }

        internal static void ReadCount(ByteBuffer buffer, FormatCode formatCode, FormatCode formatCode8, FormatCode formatCode32, out int count)
        {
            if (formatCode == formatCode8)
            {
                count = AmqpBitConverter.ReadUByte(buffer);
            }
            else if (formatCode == formatCode32)
            {
                count = (int)AmqpBitConverter.ReadUInt(buffer);
            }
            else
            {
                throw GetEncodingException(AmqpResources.GetString(AmqpResources.AmqpInvalidFormatCode, formatCode, buffer.Offset));
            }
        }

        internal static void ReadSizeAndCount(ByteBuffer buffer, FormatCode formatCode, FormatCode formatCode8, FormatCode formatCode32, out int size, out int count)
        {
            if (formatCode == formatCode8)
            {
                size = AmqpBitConverter.ReadUByte(buffer);
                count = AmqpBitConverter.ReadUByte(buffer);
            }
            else if (formatCode == formatCode32)
            {
                size = (int)AmqpBitConverter.ReadUInt(buffer);
                count = (int)AmqpBitConverter.ReadUInt(buffer);
            }
            else
            {
                throw GetEncodingException(AmqpResources.GetString(AmqpResources.AmqpInvalidFormatCode, formatCode, buffer.Offset));
            }
        }

        /// <summary>
        /// Gets the encoded size in bytes of an object.
        /// </summary>
        /// <param name="value">The object to be encoded.</param>
        /// <returns>The encoded size in bytes of an object.</returns>
        public static int GetObjectEncodeSize(object value)
        {
            if (value == null)
            {
                return FixedWidth.NullEncoded;
            }

            IAmqpSerializable serializable = value as IAmqpSerializable;
            if (serializable != null)
            {
                return serializable.EncodeSize;
            }

            EncodingBase encoding = GetEncoding(value);
            return encoding.GetObjectEncodeSize(value, false);
        }

        internal static void EncodeNull(ByteBuffer buffer)
        {
            AmqpBitConverter.WriteUByte(buffer, FormatCode.Null);
        }

        /// <summary>
        /// Encodes an object and writes the bytes to the buffer.
        /// </summary>
        /// <param name="value">The object to encode.</param>
        /// <param name="buffer">The buffer to write.</param>
        public static void EncodeObject(object value, ByteBuffer buffer)
        {
            if (value == null)
            {
                EncodeNull(buffer);
                return;
            }

            IAmqpSerializable serializable = value as IAmqpSerializable;
            if (serializable != null)
            {
                serializable.Encode(buffer);
                return;
            }

            EncodingBase encoding = GetEncoding(value);
            encoding.EncodeObject(value, false, buffer);
        }

        /// <summary>
        /// Decodes an object from the buffer.
        /// </summary>
        /// <param name="buffer">The source buffer.</param>
        /// <returns>An object.</returns>
        public static object DecodeObject(ByteBuffer buffer)
        {
            FormatCode formatCode = AmqpEncoding.ReadFormatCode(buffer);
            if (formatCode == FormatCode.Null)
            {
                return null;
            }

            return DecodeObject(buffer, formatCode);
        }

        internal static object DecodeObject(ByteBuffer buffer, FormatCode formatCode)
        {
            EncodingBase encoding;
            if (encodingsByCode.TryGetValue(formatCode, out encoding))
            {
                return encoding.DecodeObject(buffer, formatCode);
            }

            throw GetEncodingException(AmqpResources.GetString(AmqpResources.AmqpInvalidFormatCode, formatCode, buffer.Offset));
        }

        internal static AmqpException GetEncodingException(string message)
        {
            return new AmqpException(AmqpErrorCode.InvalidField, message);
        }
    }
}
